import urllib
import os
import threading
import time
import errno
from functools import partial
import weakref
import base64
import json
import socket

from socketserver import ThreadingMixIn
from http.server import SimpleHTTPRequestHandler, HTTPServer
from urllib.parse import unquote
from urllib.parse import urlparse
from urllib.parse import parse_qs

"""
HTTP Server interface
"""
class LVRequestHandler(SimpleHTTPRequestHandler, object):

    def __init__(self, viewer_weakref, *args, **kwargs):
        #Used with partial() to provide the viewer object
        try:
            self._lv = viewer_weakref
            super(LVRequestHandler, self).__init__(*args, **kwargs)
        except (IOError) as e:
            pass #Just ignore IO errors on server
            if e.errno == errno.EPIPE:
                # EPIPE error, ignore
                pass
            elif e.errno == errno.EPROTOTYPE:
                # MacOS "Protocol wrong type for socket" error, ignore
                pass
            else:
                raise e

    def serveResponse(self, data, datatype):
        try:
            #Serve provided data, with error check for SIGPIPE (broken connection)
            self.send_response(200)
            self.send_header('Content-type', datatype)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('x-colab-notebook-cache-control', 'no-cache') #Colab: disable offline access cache      
            self.end_headers()
            if data:
                self.wfile.write(data)
        #This specific error sometimes occurs on windows, ConnectionError is the base class and covers a few more
        #except (IOError,ConnectionAbortedError) as e:
        #    if isinstance(e,ConnectionAbortedError):
        except (IOError,ConnectionError) as e:
            if isinstance(e,ConnectionError):
                pass
            elif e.errno == errno.EPIPE:
                # EPIPE error, ignore
                pass
            else:
                raise e

    def do_HEAD(self):
        self.serveResponse(None, 'text/html')

    def do_POST(self):
        #Always interpret post data as commands
        #(can perform other actions too based on self.path later if we want)
        data_string = self.rfile.read(int(self.headers['Content-Length']))
        self.serveResponse(b'', 'text/plain')
        #cmds = str(data_string, 'utf-8') #python3 only
        try: #Python3
            from urllib.parse import unquote
            data_string = unquote(data_string)
        except: #Python2
            from urllib import unquote
            data_string = unquote(data_string).decode('utf8')
        cmds = str(data_string.decode('utf-8'))
        #Run viewer commands
        self._execute(cmds)

    def do_GET(self):
        lv = self._get_viewer()
        parsed = urlparse(self.path)
        query = parse_qs(parsed.query)

        def img_response():
            resp = None
            if 'width' in query and 'height' in query:
                resp = lv.jpeg(resolution=(int(query['width'][0]), int(query['height'][0])))
            elif 'width' in query:
                resp = lv.jpeg(resolution=(int(query['width'][0]), 0))
            else:
                resp = lv.jpeg()

            #Ensure the response is valid before serving
            if resp is not None:
                self.serveResponse(resp, 'image/jpeg')

        if self.path.find('image') > 0:
            img_response()

        elif self.path.find('command=') > 0:
            pos1 = self.path.find('=')
            pos2 = self.path.find('?')
            if pos2 < 0: pos2 = len(self.path)
            cmds = unquote(self.path[pos1+1:pos2])

            #Run viewer commands
            self._execute(cmds)

            #Serve image or just respond 200
            if self.path.find('icommand=') > 0:
                img_response()
            else:
                self.serveResponse(b'', 'text/plain')

        elif self.path.find('getstate') > 0:
            state = lv.app.getState()
            self.serveResponse(bytearray(state, 'utf-8'), 'text/plain; charset=utf-8')
            #self.serveResponse(bytearray(state, 'utf-8'), 'text/plain')
        elif self.path.find('connect') > 0:
            if 'url' in query:
                #Save first valid connection URL on the viewer
                url = query['url'][0]
                if len(lv._url) == 0:
                    lv._url = url
            uid = id(lv)
            self.serveResponse(bytearray(str(uid), 'utf-8'), 'text/plain; charset=utf-8')
        elif self.path.find('key=') > 0:
            pos2 = self.path.find('&')
            cmds = unquote(self.path[1:pos2])
            lv.commands('key ' + cmds, True)
            self.serveResponse(b'', 'text/plain')
        elif self.path.find('mouse=') > 0:
            pos2 = self.path.find('&')
            cmds = unquote(self.path[1:pos2])
            lv.commands('mouse ' + cmds, True)
            self.serveResponse(b'', 'text/plain')
        elif len(self.path) <= 1:
            #Root requested, returns interactive view
            w = lv.control.Window(align=None, wrapper=None)
            code = lv.control.show(True, filename="")
            self.serveResponse(bytearray(code, 'utf-8'), 'text/html; charset=utf-8')
        else:
            return SimpleHTTPRequestHandler.do_GET(self)

    #Serve files from lavavu html dir
    def translate_path(self, path):
        lv = self._get_viewer()
        if not os.path.exists(path):
            #print(' - not found in cwd')
            if path[0] == '/': path = path[1:]
            path = os.path.join(lv.htmlpath, path)
            if os.path.exists(path) and os.path.isfile(path):
                #print(' - found in htmlpath')
                return path
            else:
                #print(' - not found in htmlpath')
                return SimpleHTTPRequestHandler.translate_path(self, self.path)
        else:
            return SimpleHTTPRequestHandler.translate_path(self, path)

    #Stifle log output
    def log_message(self, format, *args):
        return

    def _get_viewer(self):
        #Get from weak reference, if deleted raise exception
        lv = self._lv()
        if not lv:
            self._closing = True
            raise(Exception("Viewer not found"))
        return lv

    def _execute(self, cmds):
        lv = self._get_viewer()

        if len(cmds) and cmds[0] == '_':
            #base64 encoded commands or JSON state
            cmds = str(base64.b64decode(cmds).decode('utf-8'))
            #cmds = str(base64.b64decode(cmds), 'utf-8')

        #Object to select can be provided in preceding angle brackets
        selobj = None
        if cmds[0] == '<':
            pos = cmds.find('>')
            selobj = lv.objects[cmds[1:pos]]
            cmds = cmds[pos+1:]

        #Execute commands via python API by preceding with '.'
        done = False
        if cmds[0] == '.':
            attr = cmds.split()[0][1:]
            pos = cmds.find(' ')
            params = cmds[pos+1:]
            if selobj:
                #Call on Object
                func = getattr(selobj, attr)
                if func and callable(func):
                    func(params)
                    done = True
            else:
                #Call on Viewer
                func = getattr(lv, attr)
                if func and callable(func):
                    func(params)
                    done = True
        elif cmds[0] == '$':
            #Requests prefixed by '$' are sent
            #from property collection controls
            #format is $ID KEY VALUE
            # - ID is the python id() of the properties object
            #    All properties collections are stored on their parent
            #    object using this id in the _collections dict
            # - KEY is the property name key to set
            # - VALUE is a json string containing the value to set
            S = cmds.split()
            target = S[0][1:]
            if target in lv._collections:
                #Get from _collections by id (weakref)
                props = lv._collections[target]()
                props[S[1]] = json.loads(S[2])
                #Check for callback - if provided, call with updated props
                func = getattr(props, 'callback')
                if func and callable(func):
                    func(props)

        #Default, call via lv.commands() scripting API
        if not done:
            if selobj:
                selobj.select()
            lv.commands(cmds)

#Optional thread per request version:
class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
    pass

"""
HTTP Server manager class
"""
class Server(threading.Thread):
    def __init__(self, viewer, port=None, ipv6=False, retries=100):
        self.host = 0
        if port is None:
            port = 8080
        self._closing = False
        #Allow viewer to be garbage collected
        self.viewer = weakref.ref(viewer)
        self.port = port
        self.retries = retries
        self.maxretries = retries
        self.ipv6 = ipv6
        super(Server, self).__init__()
        self.daemon = True #Place in background so will be closed on program exit
        self._cv = threading.Condition()

    def handle(self):
        try:
            httpd.handle_request()
        except (socket.exception) as e:
            #print(str(e))
            pass

    def run(self):
        httpd = None
        HTTPServer.allow_reuse_address = False
        try:
            # We "partially apply" our first argument to get the viewer object into LVRequestHandler
            handler = partial(LVRequestHandler, self.viewer)
            if self.ipv6:
                HTTPServer.address_family = socket.AF_INET6
                hosts = ['::', 'localhost', '::1']
                host = hosts[self.host]
                #httpd = HTTPServer((host, self.port), handler)
                httpd = ThreadingHTTPServer((host, self.port), handler)
            else:
                HTTPServer.address_family = socket.AF_INET
                hosts = ['0.0.0.0', 'localhost', '127.0.0.1']
                host = hosts[self.host]
                #httpd = HTTPServer((host, self.port), handler)
                httpd = ThreadingHTTPServer(('0.0.0.0', self.port), handler)

            #print("Server running on host %s port %s" % (host, self.port))

            #Sync with starting thread here to ensure server thread has initialised before it continues
            with self._cv:
                self._cv.notifyAll()

            # Handle requests
            #print("Using port: ", self.port)
            # A timeout is needed for server to check periodically if closing
            httpd.timeout = 0.05 #50 millisecond timeout
            while self.viewer() is not None and not self._closing:
                httpd.handle_request()

        except (Exception) as e:
            self.retries -= 1
            if self.retries < 1:
                print("Failed to start server, max retries reached")

            #Try another port
            if e.errno == errno.EADDRINUSE: #98
                self.port += 1
                #Try again
                self.run()
            elif e.errno == errno.EAFNOSUPPORT: #97 : Address family not supported by protocol
                #Try next host name/address
                self.host += 1
                if self.host > 2:
                    #Try again without ipv6?
                    if self.ipv6:
                        self.ipv6 = False
                    else:
                        self.ipv6 = True
                    self.host = 0
                #Try again
                self.run()
            else:
                print("Server start failed: ",e, e.errno, self.port)

def serve(viewer, port=None, ipv6=False, retries=100):
    s = Server(viewer, port, ipv6, retries)
    #Start the thread and wait for it to finish initialising
    with s._cv:
        s.start()
        s._cv.wait()
    return s

#Ignore SIGPIPE altogether (does not apply on windows)
import sys
if sys.platform != 'win32':
    from signal import signal, SIGPIPE, SIG_IGN
    signal(SIGPIPE, SIG_IGN)

"""
Main entry point - run server and open browser interface
"""
if __name__ == '__main__':
    import lavavu
    lv = lavavu.Viewer()
    #lv.animate(1) #Required to show viewer window and handle mouse/keyboard events there too
    lv.browser()
    lv._thread.join() #Wait for server to quit




"""
Warning! EXPERIMENTAL:
 these features and functions are under development, will have bugs,
 and may be heavily modified in the future

Tools for importing and working with point clouds
- OBJ loader, requires pywavefront
- PLY loader, requires plyfile
- LAS loader, requires laspy

These functions ported from my viz script repo here:
https://github.com/mivp/vizscripts/tree/master/pointcloud
More tools can be found there, including a ptx reader which handles transformations of multiple point clouds into one coord space

"""
import numpy
import os
import convert

def loadpointcloud(filename, subsample=1, dtype=numpy.float32, components=['x', 'y', 'z', 'red', 'green', 'blue', 'alpha']):
    """
    Attempt to load the passed point cloud file based
    Pick loader based on extension

    Parameters
    ----------
    filename, str
        Full path to file to load
    subsample, int
        Subsample factor, sample every Nth point
    dtype, numpy.dtype
        Type of data
    components, list
        List of components per point

    Returns
    -------
    vertices : array
        numpy array x,y,z
    colours : array
        numpy array r,g,b or r,g,b,a
    """
    fn, ext = os.path.splitext(filename)
    ext = ext.lower()
    if ext == '.xyz':
        #Loop over lines in input file
        count = 0
        V = []
        C = []
        with open(filename, 'r') as file:
            for line in file:
                #Subsample?
                count += 1
                if subsample > 1 and count % subsample != 1: continue

                if count % 10000 == 0:
                  print(count)
                  sys.stdout.flush()

                #Read particle position
                data = re.split(r'[,;\s]+', line.rstrip())
                if len(data) < 3: continue
                x = float(data[0])
                y = float(data[1])
                z = float(data[2])

                V.append((x, y, z))

                #R,G,B[,A] colour if provided
                if len(data) >= 7:
                    C.append((int(data[3]), int(data[4]), int(data[5]), int(data[6])))
                elif len(data) == 6:
                    C.append((int(data[3]), int(data[4]), int(data[5]), 255))

        V = numpy.array(V)
        C = numpy.array(C)

    elif ext == '.xyzb':
        inbytes = os.path.getsize(filename)
        #All floats
        #num = inbytes / (4 * 6) # x y z r g b
        #All floats, rgba
        #num = inbytes / (4 * 7) # x y z r g b a
        #All doubles, rgba
        #num = inbytes / (8 * 7) # x y z r g b a
        #Float vert, byte colour
        #num = inbytes / (4 * 3 + 4) # x y z r g b a
        sz = numpy.dtype(dtype).itemsize
        num = inbytes // (sz * len(components))

        with open(filename, 'rb') as infile:
            print("Importing " + str(num//subsample) + " points")
            arr = numpy.fromfile(infile, dtype='float64')
            infile.close()

            arr = arr.astype(dtype).reshape(len(arr)//len(components), len(components))
            if subsample > 1:
                arr = arr[::subsample]

            #Convert float rgba to uint32 colour
            colours = None
            if len(components) > 6:
                colours = arr[:,3:len(components)]
                colours *= 255
                colours = colours.astype('uint8').ravel()
                colours = colours.view(dtype='uint32')

            return (arr[:,0:3], colours)

    elif ext == '.obj':
        print("Loading OBJ")
        import pywavefront
        scene = pywavefront.Wavefront(filename)
        V = numpy.array(scene.vertices)

        #SubSample
        if subsample > 1:
            V = V[::subsample]
            print("Subsampled:",V.shape)

        verts = V[:,0:3]
        colours = V[:,3:] * 255

        #Load positions and RGB
        print("Creating visualisation ")
        return (verts, colours)

    elif ext == '.ply':
        print("Loading PLY")
        from plyfile import PlyData, PlyElement       
        plydata = PlyData.read(filename)
        if plydata:
            x = plydata['vertex']['x']
            y = plydata['vertex']['y']
            z = plydata['vertex']['z']
            V = numpy.vstack((x,y,z)).reshape([3, -1]).transpose()

            C = convert._get_PLY_colours(plydata.elements[0])

            #SubSample
            if subsample > 1:
                V = V[::subsample]
                if C is not None:
                    C = C[::subsample]

            return (V, C)

    elif ext == '.las' or ext == '.laz':
        print("Loading LAS")
        import laspy
        infile = laspy.file.File(filename, mode="r")
        '''
        print(infile)
        print(infile.point_format)
        for spec in infile.point_format:
           print(spec.name)
        print(infile.header)
        print(infile.header.dataformat_id)

        print(infile.header.offset)
        print(infile.header.scale)
        '''

        #Grab all of the points from the file.
        #point_records = infile.points
        #print(point_records)
        #print(point_records.shape)

        #Convert colours from short to uchar
        if infile.red.dtype == numpy.uint16:
            R = (infile.red / 255).astype(numpy.uint8)
            G = (infile.green / 255).astype(numpy.uint8)
            B = (infile.blue / 255).astype(numpy.uint8)
        else:
            R = infile.red
            G = infile.green
            B = infile.blue

        ss = subsample
        if ss > 1:
            V = numpy.array([infile.x[::ss], infile.y[::ss], infile.z[::ss]])
            C = numpy.array([R[::ss],G[::ss],B[::ss]])
            return (V, C)
        else:
            V = numpy.array([infile.x, infile.y, infile.z])
            C = numpy.array([R,G,B])
            return (V, C)

    else:
        print("Unknown point cloud format, extension: ", ext)




import os
import sys
from setuptools import setup
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
from setuptools import Extension
from setuptools import find_packages
try:
    from setuptools import _distutils as distutils
    from setuptools._distutils import sysconfig
    from setuptools._distutils import ccompiler
except ImportError:
    import distutils
    from distutils import sysconfig
    from distutils import ccompiler
import subprocess
from multiprocessing import cpu_count
from ctypes.util import find_library
import platform
import glob
import shutil

#Current version
#(must be of the form X.Y.Z to trigger wheel builds)
version = "1.8.60"

"""
To release a new verison:

    1) Edit the version number above, then run below command (in a fresh checkout!).
    This will rebuild lib and docs in release mode with the new version
    (may need to reset notebooks to documentation versions)

    ```
    #Get clean checkout and setup nested docs repo
    git clone git@github.com:lavavu/LavaVu.git
    cd LavaVu
    rm -rf docs
    git clone git@github.com:lavavu/Documentation.git docs
    git checkout HEAD -- docs
    cd docs/src
    pip install -r requirements.txt
    ```
    (lib dependencies for debian/ubuntu: `sudo apt install pandoc emscripten`)

    >>> python setup.py check

    2) Check the docs, and make any further commits necessary to get them in order,
       and run above command again until docs are ready.

    Now ready for release, run this to commit and push changes to docs,
    then tag, commit and push new version.

    >>> python setup.py release

NOTE:
    To use particular libraries, set LV_LIB_DIRS and LV_INC_DIRS on command line before running, eg:
    LV_LIB_DIRS=${MESA_PATH/lib/x86_64-linux-gnu LV_INC_DIRS=${MESA_PATH}/include python setup.py install

"""

def write_version():
    """
    Writes version info to version.cpp
    """
    content = ""
    try:
        with open('src/version.cpp', 'r') as f:
            content = f.read()
    except:
        pass
    if not version in content:
        with open('src/version.cpp', 'w') as f:
            print("Writing new version: " + version)
            f.write('#include "version.h"\nconst std::string version = "%s";\n' % version)
    else:
        print("Version matches: " + version)

#Run with "new" to prepare for a new version after changing above
if sys.argv[-1] == 'check':
    #Check docs repo in sync first to avoid merge errors on release step
    wd = os.getcwd()
    os.chdir("docs")
    os.system("git pull")
    os.chdir("src")
    os.system("pip install -r requirements.txt")
    os.chdir(wd)
    write_version()
    os.system("LV_RELEASE=1 make clean")
    os.system("LV_RELEASE=1 make -j$(nproc)")
    os.system("LV_RELEASE=1 make docs -B")
    os.system("LV_RELEASE=1 make clean")
    os.system("LV_RELEASE=1 make emscripten -j$(nproc)")
    os.system("cd lavavu; python amalgamate.py; cd ..")
    import webbrowser
    webbrowser.open("docs/index.html")
    sys.exit()
#Run with "tag" arg to create a release tag
if sys.argv[-1] == 'release':
    #Commit and push docs
    wd = os.getcwd()
    os.chdir("docs")
    os.system("git pull")
    os.system("git add -u")
    os.system('git commit -m "Docs rebuilt for release"')
    os.system("git push")
    #Commit main repo
    os.chdir(wd)
    os.system("git add -u")
    os.system('git commit -m "Release version {0}"'.format(version))

    #Tag and push
    os.system("git tag -a {0} -m 'version {0}'".format(version))
    os.system("git push origin %s" % version)
    os.system("git push")
    sys.exit()

#Run with "publish" arg to upload the release
if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist")
    os.system("twine upload dist/lavavu-%s.tar.gz" % version)
    sys.exit()

# Just output the wheel name for current arch + environment
if sys.argv[-1] == 'wheelname':
    # https://stackoverflow.com/a/60644659/866759
    from setuptools.dist import Distribution
    lavavu = Extension('lavavu', [])
    dist = Distribution(attrs={'name': 'lavavu', 'version': version, 'ext_modules': [lavavu]})
    bdist_wheel_cmd = dist.get_command_obj('bdist_wheel')
    bdist_wheel_cmd.ensure_finalized()
    distname = bdist_wheel_cmd.wheel_dist_name
    tag = '-'.join(bdist_wheel_cmd.get_tag())
    wheel_name = f'{distname}-{tag}.whl'
    print(wheel_name)
    sys.exit()

#Get extra lib and include dirs
inc_dirs = []
lib_dirs = []
if 'LV_LIB_DIRS' in os.environ:
    lib_dirs += os.environ['LV_LIB_DIRS'].split(':')
    #Add dirs to ld_library_path
    old = os.environ.get("LD_LIBRARY_PATH")
    if old:
        os.environ["LD_LIBRARY_PATH"] = old + ":" + os.environ['LV_LIB_DIRS']
    else:
        os.environ["LD_LIBRARY_PATH"] = os.environ['LV_LIB_DIRS']
if 'LV_INC_DIRS' in os.environ:
    inc_dirs += os.environ['LV_INC_DIRS'].split(':')

def build_sqlite3(sqlite3_path):
    """Builds sqlite3 from included submodule
    """
    compiler = ccompiler.new_compiler()
    #assert isinstance(compiler, ccompiler.CCompiler)
    sysconfig.customize_compiler(compiler)

    #Add any include dirs
    compiler.add_include_dir(sqlite3_path)
    compiler.define_macro('SQLITE_ENABLE_DESERIALIZE', '1')

    try:
        obj = compiler.compile([os.path.join(sqlite3_path, 'sqlite3.c')], 'build')
        return obj
    except Exception:
        print('sqlite3 compile error')
        raise

#From https://stackoverflow.com/a/28949827/866759
def check_libraries(libraries, headers, extra_lib_dirs=[], extra_inc_dirs=[]):
    """check if the C module can be built by trying to compile a small
    program against the passed library with passed headers"""

    import tempfile
    # write a temporary .c file to compile
    c_code = "int main(int argc, char* argv[]) { return 0; }"
    #Add headers
    for header in headers:
        c_code = "#include <" + header + ">\n" + c_code

    tmp_dir = tempfile.mkdtemp(prefix = 'tmp_comp__')
    bin_file_name = os.path.join(tmp_dir, 'test_comp')
    file_name = bin_file_name + '.c'
    with open(file_name, 'w') as fp:
        fp.write(c_code)

    # and try to compile it
    compiler = ccompiler.new_compiler()
    #assert isinstance(compiler, ccompiler.CCompiler)
    sysconfig.customize_compiler(compiler)

    #Add any extra library or include dirs
    for l in lib_dirs + extra_lib_dirs:
        compiler.add_library_dir(l)
    for p in inc_dirs + extra_inc_dirs:
        compiler.add_include_dir(p)


    try:
        compiler.link_executable(
            compiler.compile([file_name]),
            bin_file_name,
            libraries=libraries,
        )
    except (Exception) as e:
        print('Libraries ' + str(libraries) + ' test build error', e)
        ret_val = False
    else:
        print('Libraries ' + str(libraries) + ' found and passed compile test')
        ret_val = True
    shutil.rmtree(tmp_dir)
    return ret_val

if __name__ == "__main__":
    print("Running setup with args:", sys.argv)
    #Update version.cpp
    write_version()

    sqlite3_path = os.path.join('src', 'sqlite3')
    if not os.path.exists(os.path.join(sqlite3_path, 'sqlite3.c')):
        #Attempt to get sqlite3 source submodule if not checked out
        os.system("git submodule update --init")

    _debug = False
    if "develop" in sys.argv:
        _debug = True
    srcs = [os.path.join('src', 'LavaVuPython_wrap.cxx')]
    srcs += glob.glob(os.path.join('src', '*.cpp'))
    srcs += glob.glob(os.path.join('src', 'Main', '*.cpp'))
    srcs += glob.glob(os.path.join('src', 'jpeg', '*.cpp'))
    defines = [('USE_FONTS', '1')]
    cflags = []
    libs = []
    ldflags = []
    rt_lib_dirs = []
    extra_objects = []
    try:
        import numpy
    except:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'numpy>=1.11.0'])
        try:
            import numpy
        except:
            subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'numpy>=1.11.0'])
            import numpy
    inc_dirs += [numpy.get_include()]

    P = platform.system()
    if _debug:
        print("DEBUG BUILD ENABLED")
        defines += [('DEBUG', '1')]
        if P == 'Windows':
            cflags += ['/Zi']
            ldflags += ['/DEBUG']
        else:
            cflags += ['-g', '-O0']

    #Parallel build
    num_jobs = 1
    if not 'arm' in platform.machine():
        try:
            num_jobs = os.cpu_count()
        except AttributeError:
            import multiprocessing
            num_jobs = multiprocessing.cpu_count()

    if num_jobs > 1:
        try:
            from numpy.distutils.ccompiler import CCompiler_compile
            from setuptools._distutils import ccompiler
            ccompiler.CCompiler.compile = CCompiler_compile
            os.environ['NPY_NUM_BUILD_JOBS'] = str(num_jobs)
            print('Using parallel build, jobs: ', num_jobs)
        except ImportError:
            print("Numpy not found, parallel compile not available")

    #OS Specific
    if P == 'Windows':
        #Windows - include all dependencies

        #VC++ can handle C or C++ without issues clang has,
        #so just throw sqlite source in with the rest
        inc_dirs += [sqlite3_path]
        srcs += [os.path.join(sqlite3_path, 'sqlite3.c')]

        #32 or 64 bit python interpreter?
        if sys.maxsize > 2**32:
            arc = '64'
        else:
            arc = '32'

        #Download, extract and install binary libraries
        win_dlls = []
        win_libs = []
        try:
            sys.path.append('lavavu')
            import vutils
            import zipfile
            src = "https://codeload.github.com/lavavu/windows/zip/refs/heads/main"
            fnz = "windows-main.zip"
            vutils.download(src, fnz)
            with zipfile.ZipFile(fnz, 'r') as zip_ref:
                zip_ref.extractall('.')
            LIBS = 'lib' + arc
            # - Lib and dll files - just grab by current location
            win_dlls = glob.glob(os.path.join('windows-main', LIBS, '*.dll'))
            win_libs = glob.glob(os.path.join('windows-main', LIBS, '*.lib'))
            #If we got this far, enable video, tiff
            defines += [('HAVE_LIBAVCODEC', '1'), ('HAVE_SWSCALE', '1')]
            defines += [('HAVE_LIBTIFF', '1')]

        except (Exception) as e:
            print("windows binary libraries download/extract failed", str(e))
            win_dlls = []
            win_libs = []
            pass

        srcs += [os.path.join('src', 'png', 'lodepng.cpp')]
        srcs += [os.path.join('src', 'miniz', 'miniz.c')]
        defines += [('HAVE_GLFW', '1'), ('SQLITE_ENABLE_DESERIALIZE', '1')]
        #defines += [('HAVE_LIBPNG', '1')]
        inc_dirs += [os.path.join('windows-main', 'inc')]
        lib_dirs += [os.path.join('windows-main', LIBS)]
        ldflags += ['/LIBPATH:' + os.path.join('windows-main', LIBS)]
        from pathlib import Path
        libs = ['opengl32'] + [Path(l).stem for l in win_libs]
        #Copy dlls into ./lavavu so can be found by package_data
        for d in win_dlls:
            shutil.copy(d, 'lavavu')
        #print(libs)
        #print(inc_dirs)
        #print(lib_dirs)
    else:
        #POSIX only - find external dependencies
        cflags += ['-std=c++0x']
        #Build sqlite3
        #Don't compile when just running 'setup.py egg_info'
        if len(sys.argv) < 2 or sys.argv[1] == 'egg_info':
            sqlite3 = 'build/src/sqlite3/sqlite3.o'
        else:
            sqlite3 = build_sqlite3(sqlite3_path)
        inc_dirs += [sqlite3_path]
        extra_objects = sqlite3

        # Optional external libraries - check if installed
        if find_library('png') and check_libraries(['png', 'z'], ['png.h', 'zlib.h']):
            defines += [('HAVE_LIBPNG', 1), ('USE_ZLIB', '1')]
            libs += ['png', 'z']
        else:
            srcs += ['src/png/lodepng.cpp']
            if find_library('z') and check_libraries(['z'], ['zlib.h']):
                defines += [('USE_ZLIB', '1')]
                libs += ['z']
            else:
                srcs += ['src/miniz/miniz.c']

        if find_library('tiff') and check_libraries(['tiff'], ['tiffio.h']):
            defines += [('HAVE_LIBTIFF', 1)]
            libs += ['tiff']

        if (find_library('avcodec') and find_library('avformat') and find_library('avutil')
            and check_libraries(['avcodec', 'avformat', 'avutil'],
                ['libavformat/avformat.h', 'libavcodec/avcodec.h', 'libavutil/mathematics.h',
                 'libavutil/imgutils.h'])):
            defines += [('HAVE_LIBAVCODEC', 1)]
            libs += ['avcodec', 'avformat', 'avutil']
            if find_library('swscale') and check_libraries(['swscale'], ['libswscale/swscale.h']):
                defines += [('HAVE_SWSCALE', 1)]
                libs += ['swscale']

        if P == 'Linux':
            #Linux GLFW, X11, EGL or OSMesa
            #To force EGL or OSMesa, set LV_EGL=1 or LV_OSMESA=1
            if ("LV_EGL" in os.environ and find_library('OpenGL') and find_library('EGL')
            and check_libraries(['OpenGL', 'EGL'], ['GL/gl.h', 'EGL/egl.h'])):
                #EGL for offscreen OpenGL without X11/GLX - works only with NVidia currently
                defines += [('HAVE_EGL', '1')]
                libs += ['OpenGL', 'EGL']
            elif ("LV_OSMESA" in os.environ and find_library('OSMesa')
            and check_libraries(['OSMesa'], ['GL/osmesa.h'])):
                #OSMesa for software rendered offscreen OpenGL
                defines += [('HAVE_OSMESA', '1')]
                libs += ['OSMesa']
            elif ("LV_GLFW" in os.environ and find_library('glfw')
            and check_libraries(['glfw'], ['GLFW/glfw3.h'])):
                #Use GLFW
                defines += [('HAVE_GLFW', '1')]
                libs += ['GL', 'glfw']
            else:
                #Fallback - X11
                defines += [('HAVE_X11', '1')]
                libs += ['GL', 'X11']

            #Runtime libraries: set rpath: $ORIGIN includes current dir in search
            rt_lib_dirs += ['$ORIGIN'] + lib_dirs

        elif P == 'Darwin':
            #Mac OS X with Cocoa + CGL
            #srcs += ['src/Main/CocoaViewer.mm']
            #This hack is because setuptools can't handle .mm extension
            shutil.copyfile('src/Main/CocoaViewer.mm', 'src/Main/CocoaViewer.m')
            srcs += ['src/Main/CocoaViewer.m']
            cflags += ['-ObjC++'] #Now have to tell compiler it's objective c++ as .m file indicates objective c
            defines += [('HAVE_CGL', '1'), ('GL_SILENCE_DEPRECATION','1')]
            cflags += ['-undefined suppress', '-flat_namespace'] #Swig, necessary?
            cflags += ['-Wno-unknown-warning-option', '-Wno-c++14-extensions', '-Wno-shift-negative-value']
            cflags += ['-FCocoa', '-FOpenGL', '-stdlib=libc++']
            libs += ['c++', 'objc']
            #Can't use extra_link_args because they are appended by setuptools but frameworks must come first
            os.environ['LDFLAGS'] = '-framework Cocoa -framework Quartz -framework OpenGL'
            #Runtime library dirs doesn't work on mac, so set rpath manually
            for l in lib_dirs:
                ldflags.append('-Wl,-rpath,'+l)

        #Other posix libs
        libs += ['dl', 'pthread', 'm']

    lv = Extension('lavavu._LavaVuPython',
                    define_macros = defines,
                    include_dirs = inc_dirs,
                    libraries = libs,
                    library_dirs = lib_dirs,
                    runtime_library_dirs = rt_lib_dirs,
                    extra_compile_args = cflags,
                    extra_link_args = ldflags,
                    extra_objects=extra_objects,
                    sources = srcs)

    #For some reason, jupyter_server_proxy requirement breaks conda build
    requirements = ['numpy>=1.11', 'aiohttp', 'jupyter_server_proxy']
    if 'CONDA_BUILD' in os.environ:
        del requirements[-1]

    #Binary package data for wheels
    #Package_data works for wheels(binary) only - so add everything we need for the wheels here
    #(MANIFEST.in contents will be included with sdist only unless include_package_data is enabled)
    package_data = {'lavavu': ['font.bin', 'dict.json', 'shaders/*', 'html/*']}
    if P == 'Windows':
        package_data['lavavu'] += ['*.dll']

    dist_name = "lavavu"
    #Rename package so can be distributed as an alternative (eg: lavavu-osmesa)
    if 'LV_PACKAGE' in os.environ:
        dist_name = os.environ['LV_PACKAGE']

    setup(name = dist_name,
          author            = "Owen Kaluza",
          author_email      = "owen.kaluza@monash.edu",
          url               = "https://github.com/lavavu/LavaVu",
          version           = version,
          license           = "LGPL-3",
          description       = "Python interface to LavaVu OpenGL 3D scientific visualisation utilities",
          long_description  = 'See https://github.com/lavavu/LavaVu for more info',
          packages          = ['lavavu'],
          package_dir       = {'lavavu': 'lavavu'},
          package_data      = package_data,
          install_requires  = requirements,
          platforms         = ['any'],
          entry_points      = {
              'gui_scripts': [
                  'LV = lavavu.__main__:main',
                  'LavaVu = lavavu.__main__:main'
              ]
          },
          #Use below to include files from MANIFEST.in or specify package_data, not both
          #include_package_data = True,
          classifiers = [
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS',
            'Operating System :: Microsoft :: Windows',
            'Environment :: X11 Applications',
            'Environment :: MacOS X :: Cocoa',
            'Environment :: Win32 (MS Windows)',
            'Topic :: Multimedia :: Graphics :: 3D Rendering',
            'Topic :: Scientific/Engineering :: Visualization',
            'Development Status :: 5 - Production/Stable',
            'Programming Language :: C++',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Framework :: Jupyter',
            'Framework :: IPython',
          ],
          ext_modules = [lv]
          )




from __future__ import print_function
"""
Warning! EXPERIMENTAL:
 these features and functions are under development, will have bugs,
 and may be heavily modified in the future

Tools for converting between 3D data types

- Points to Volume
- Triangles to OBJ file
"""
import numpy
import os
import sys

def try_import(module_name, second_try=False):
    """
    Attempts to import a module, then runs pip install if not found and attempts again
    """
    import importlib
    try:
        m = importlib.import_module(module_name)
        globals()[module_name] = m
        return m
    except (ImportError) as e:
        if not second_try:
            import subprocess
            subprocess.check_call([sys.executable, "-m", "pip", "install", module_name])
            return try_import(module_name, True)
        else:
            raise(e)

def min_max_range(verts):
    """
    Get bounding box from a list of vertices
    returns (min, max, range)
    """
    vmin = numpy.min(verts, axis=0)
    vmax = numpy.max(verts, axis=0)
    vrange = vmax - vmin
    print("Bounding box ", (vmin, vmax), "Range ", vrange)
    return (vmin, vmax, vrange)

def default_sample_grid(vrange, res=8):
    """Calculate sample grid fine enough to capture point details
    resolution of smallest dimension will be minres elements
    """
    #If provided a full resolution, use that, otherwise will be interpreted as min res
    if isinstance(res, (list,tuple)):
        if len(list(res)) == 3:
            return list(res)
        res = res[0]
    #Use bounding box range min
    minr = numpy.min(vrange)
    #res parameter is minimum resolution
    factor = float(res) / float(minr)
    RES = [int(factor*(vrange[0])), int(factor*(vrange[1])), int(factor*(vrange[2]))]   
    print("Sample grid RES:",RES)
    return RES

def points_to_volume(verts, weights=None, res=8, kdtree=False, normed=True, clamp=None, boundingbox=None):
    """
    Convert object vertices to a volume by interpolating points to a grid

    - Vertex data only is used, so treated as a point cloud
    - Can also be used to sample an irregular grid to a regular so volume render matches the grid dimensions
    - Result is a density field that can be volume rendered

    Default is to use numpy.histogramdd method, pass kdtree=True to use scipy.spatial.KDTree

    TODO: support colour data too, converted density field becomes alpha channel (actually, use weights)

    Returns
    -------
    values: numpy array of float32
        The converted density field
    boundingbox : 2,3
        the minimum 3d vertex of the bounding box
        the maximum 3d vertex of the bounding box

    """
    if kdtree:
        return points_to_volume_tree(verts, weights, res)
    else:
        return points_to_volume_histogram(verts, weights, res, normed, clamp, boundingbox)

def points_to_volume_histogram(verts, weights, res=8, normed=True, clamp=None, boundingbox=None):
    """
    Using numpy.histogramdd to create 3d histogram volume grid
    (Easily the fastest, but less control over output)
    """
    #Reshape to 3d vertices
    verts = verts.reshape((-1,3))

    #Get bounding box of swarm
    vmin, vmax, vrange = min_max_range(verts)

    #Minimum resolution to get ok sampling
    if boundingbox is not None:
        vmin = boundingbox[0]
        vmax = boundingbox[1]
        vrange = numpy.array(boundingbox[1]) - numpy.array(boundingbox[0])
    RES = default_sample_grid(vrange, res)

    if weights is not None:
        weights = weights.ravel()
        #Normalise (should be done by histogramdd anyway if normed=True)
        weights = (weights - weights.min()) / (weights.max() - weights.min())
        #print(verts.shape, weights.shape, weights.min(), weights.max())

    #H, edges = numpy.histogramdd(verts, bins=RES)
    if boundingbox is None:
        H, edges = numpy.histogramdd(verts, weights=weights, bins=RES, normed=normed) #density=True for newer numpy
    else:
        rg = ((vmin[0], vmax[0]), (vmin[1], vmax[1]), (vmin[2], vmax[2])) #provide bounding box as range
        H, edges = numpy.histogramdd(verts, weights=weights, bins=RES, range=rg, normed=normed) #density=True for newer numpy

    #Reverse ordering X,Y,Z to Z,Y,X for volume data
    values = H.transpose()

    #Clamp [0,0.1] - optional (for a more binary output, points vs no points)
    if clamp is not None:
        values = numpy.clip(values, a_min=clamp[0], a_max=clamp[1])
    
    #Normalise [0,1] if using probability density
    if normed:
        values = values / numpy.max(values)
    
    return (values, vmin, vmax)

def points_to_volume_tree(verts, res=8):
    """
    Using scipy.spatial.KDTree to find nearest points on grid

    Much slower, but more control

    TODO: control parameters
    """
    #Reshape to 3d vertices
    verts = verts.reshape((-1,3))

    #Get bounding box of swarm
    lmin, lmax, lrange = min_max_range(verts)
    print("Volume bounds: ",lmin.tolist(), lmax.tolist(), lrange.tolist())
    
    #Push out the edges a bit, will create a smoother boundary when we filter
    #lmin -= 0.1*lrange
    #lmax += 0.1*lrange
    values = numpy.full(shape=(verts.shape[0]), dtype=numpy.float32, fill_value=1.0)

    #Minimum resolution to get ok sampling
    RES = default_sample_grid(lrange, res)

    #Push out the edges a bit, will create a smoother boundary when we filter
    cell = lrange / RES #Cell size
    print("CELL:",cell)
    lmin -= 2.0*cell
    lmax += 2.0*cell

    x = numpy.linspace(lmin[0], lmax[0] , RES[0])
    y = numpy.linspace(lmin[1], lmax[1] , RES[1])
    z = numpy.linspace(lmin[2], lmax[2] , RES[2])
    print(x.shape, y.shape, z.shape)
    Z, Y, X = numpy.meshgrid(z, y, x, indexing='ij') #ij=matrix indexing, xy=cartesian
    XYZ = numpy.vstack((X,Y,Z)).reshape([3, -1]).T

    print(lmin,lmax, XYZ.shape)

    #KDtree for finding nearest neighbour points
    from scipy import spatial
    #sys.setrecursionlimit(1000)
    #tree = spatial.KDTree(XYZ)
    print("Building KDTree")
    tree = spatial.cKDTree(XYZ)

    #Outside distance to apply to grid points
    MAXDIST = max(lrange) / max(RES) #Max cell size diagonal
    print("Tree query, maxdist:",MAXDIST,max(lrange))

    #Query all points, result is tuple: distances, indices
    distances, indices = tree.query(verts, k=1) #Just get a single nearest neighbour for now

    print("Calculate distances")
    #Convert distances to [0,1] where 1=on grid and <= 0 is outside max range
    distances = (MAXDIST - distances) / MAXDIST
    print("Zero out of range distances")
    distances *= (distances>0) #Zero negative elements by multiplication in-place

    #Add the distances to the values at their nearest grid point indices
    print("Compose distance field")
    values = numpy.zeros(shape=(XYZ.shape[0]), dtype=numpy.float32)
    values[indices] += distances
    #Clip value max
    print("Clip distance field")
    values = values.clip(max=1.0)
    
    #Reshape to actual grid dims Z,Y,X... (not required but allows slicing)
    XYZ = XYZ.reshape(RES[::-1] + [3])
    values = values.reshape(RES[::-1])
    print(XYZ.shape, values.shape)
    
    return (values, lmin, lmax)


def points_to_volume_3D(vol, objects, res=8, kdtree=False, blur=0, pad=None, normed=True, clamp=None, weights=None):
    """
    Interpolate points to grid and load into passed volume object

    Given a list of objects and a volume object, convert a point cloud
    from another object (or objects - list is supported) into a volume using
    points_to_volume()
    """
    lv = vol.parent #Get the viewer from passed object
    #lv.hide(objects) #Hide the converted objects

    #Get vertices from lavavu objects
    if weights:
        pverts, bb_all, weights = lv.get_all_vertices(objects, weights)
    else:
        pverts, bb_all = lv.get_all_vertices(objects)

    #blur = False
    #Use bounding box of full model?
    #vmin, vmax, vrange = min_max_range([lv["min"], lv["max"]])
    #bb_all == (vmin, vmax)
    vdata, vmin, vmax = points_to_volume(pverts, weights, res, kdtree, normed, clamp, bb_all)

    if blur > 0:
        if pad == None:
            pad = int(blur*2)
        if pad > 0:
            print("Pad edges before blur", pad)
            vdata = numpy.pad(vdata, pad, mode='constant')
        print("Filter/blur distance field, sigma=%d" % blur)
        from scipy.ndimage.filters import gaussian_filter
        values = gaussian_filter(vdata, sigma=blur) #, mode='nearest')
    else:
        values = vdata #Need extra space at edges to use blur, so skip, or use pad() 
    
    print(numpy.min(values), numpy.max(values))
    print(numpy.min(values), numpy.max(values))
    
    vol.values(values)
    vol.vertices((vmin, vmax))

def points_to_volume_4D(vol, objects, res=8, kdtree=False, blur=0, pad=None, normed=True, clamp=None, weights=None):
    """
    Interpolate points to grid at each timestep

    Given a list of objects and a volume object, convert a time-varying point cloud
    from another object (or objects - list is supported) into a volume using
    points_to_volume_3D()
    """
    lv = vol.parent #Get the viewer from passed object

    for step in lv.steps:
        print("TIMESTEP:",step)
        lv.timestep(step)
        #TODO: timestep data not changing!!!
        print(lv.step)
        lv.display((200,150))
        
        points_to_volume_3D(vol, objects, res, kdtree, blur, pad, normed, clamp, weights)

def colour2rgb(c):
    return [c & 255, (c >> 8) & 255, (c >> 16) & 255]

def colour2hex(rgb):
    def padhex2(i):
        s = hex(int(i))
        return s[2:].zfill(2)
    return "#" + padhex2(rgb[0]) + padhex2(rgb[1]) + padhex2(rgb[2])

def _get_objects(source):
    """
    Returns a list of objects

    If source is lavavu.Viewer() list contains all objects
    If source is lavavu.Object() list contains that single object
    """
    if source.__class__.__name__ == 'Viewer':
        return source.objects.list
    elif not isinstance(source, list):
        return [source]
    else:
        return source

def export_OBJ(filepath, source, verbose=False, vertexcolours=True):
    """
    Export given object(s) to an OBJ file
    Supports only triangle mesh object data

    If source is lavavu.Viewer() exports all objects
    If source is lavavu.Object() exports single object

    Set vertexcolours to support writing R,G,B values with vertices
    (not part of OBJ standard but supported in some software)
    """
    mtlfilename = os.path.splitext(filepath)[0] + '.mtl'
    objects = _get_objects(source)
    with open(filepath, 'w') as f, open(mtlfilename, 'w') as m:
        f.write("# OBJ file\n")
        offset = 1
        for obj in objects:
            if obj["visible"]:
                f.write("g %s\n" % obj.name)
                offset = _write_OBJ(f, m, filepath, obj, offset, verbose, vertexcolours)

def _write_MTL(m, name, texture=None, diffuse=[1.0, 1.0, 1.0], ambient=None, specular=None, opacity=1.0, illum=None):
    #http://paulbourke.net/dataformats/mtl/
    #print("Writing MTL ", texture, diffuse, opacity, name)
    mtl = "newmtl %s\n" % name
    mtl += "Kd %.06f %.06ff %.06f\n" % (diffuse[0], diffuse[1], diffuse[2])
    if ambient:
        mtl += "Ka %.06f %.06ff %.06f\n" % (ambient[0], ambient[1], ambient[2])
    if specular:
        if illum is None:
            illum = 2 #Highlight on
        mtl += "Ks %.06f %.06ff %.06f\n" % (specular[0], specular[1], specular[2])
        if len(specular) > 3:
            mtl += "Ns %.06f\n" % specular[3]
    mtl += "d %f\n" % (1.0 - opacity) #Dissolve: inverse of opacity
    if illum is None:
        illum = 1 #Default=1 = colour on, ambient on
    mtl += "illum %d\n" % illum

    if texture:
        mtl += "map_Kd %s\n" % texture

    mtl += '\n'
    m.write(mtl)
    mtl_line = "usemtl %s\n" % name
    return mtl_line

def _write_OBJ(f, m, filepath, obj, offset=1, verbose=False, vertexcolours=True):
    mtl_line = ""
    cmaptexcoords = []
    colourdict = None
    import re
    name = re.sub(r"\s+", "", obj["name"])
    colourcount = sum([len(c) for c in obj.data.colours])
    #print(f"{colourcount=}")
    if m and obj["texture"]:
        fn = obj["texture"]
        if fn == 'colourmap':
            fn = 'texture.png'
        #Write palette.png
        obj.parent.palette('image', 'texture')
        if verbose:
            print("Writing texture mtl ", fn)
        mtl_line = _write_MTL(m, name, texture=fn, opacity=obj["opacity"])

    elif m and colourcount > 0 and not vertexcolours:
        #Define material for each (slow for high colour/vertex count - could sort faces by material before writing?)
        if colourcount < 10000:
            colourdict = {}
            #SLOW!

            #Get unique https://stackoverflow.com/a/33197029/866759
            allcolours = numpy.concatenate(obj.data.colours)
            unique = numpy.unique(allcolours)

            for c in unique:
                rgb = colour2rgb(c)
                cs = colour2hex(rgb)
                colourdict[c] = (cs,rgb)
            if verbose:
                print("Writing mtl lib (colour list), unique colours:",len(unique))
            for c in colourdict:
                cs, rgb = colourdict[c]
                mtl_line = _write_MTL(m, cs, diffuse=[rgb[0]/255.0, rgb[1]/255.0, rgb[2]/255.0], opacity=obj["opacity"])
        else:
            #Support this full RGB palette?
            #TODO: generate texcoord data using formula below
            #18bit RGB = 262144 colours
            #Top-left = (0,0,0)
            #Bottom-right = (255,255,255)
            #(8x8 tiles of 64x64)
            #G - slow, Z index = G
            #R - midd, Y axis = R
            #B - fast, X axis = B
            #X = B//4
            #Y = R//4
            #Z = G//4
            #zx = Z % 8
            #zy = Z // 8
            #x = zx * 64 + X
            #y = zy * 64 + Y
            #u = x / 512
            #v = y / 512

            #https://upload.wikimedia.org/wikipedia/commons/3/34/RGB_18bits_palette.png
            palimg = "iVBORw0KGgoAAAANSUhEUgAAAgAAAAIACAIAAAB7GkOtAAAABGdBTUEAANjr9RwUqgAAACBjSFJNAACH"\
                     "CgAAjAoAAPYWAACEzwAAczsAAOxVAAA6lwAAHU1girkoAAAGc0lEQVR42u3dQYoDMQxFwTZY93Duf8mQ"\
                     "A8hg9yIIVTHbgU9vHsrG8/mZL/7iv/8+n2G//fbbb//V/vkA0JAAAAgAAAIAgAAAIAAACAAAAgCAAAAg"\
                     "AAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgDARQDCRwBwAQAgAAAIAAACAIAAACAAAAgAAAIAgAAA"\
                     "IAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAQB4Aj8IDuAAAEAAABAAAAQBAAAAQAAAEAAABAEAA"\
                     "ABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAGATAI/CA/QMwBrVC2a//fbbb//dfj8BAfS8AAQA"\
                     "QAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAADOAxA+AoALAAAB"\
                     "AEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAA8gB4FB7ABQCA"\
                     "AAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAJsAeBQeoGcA"\
                     "1qheMPvt77s/fH/7X+33ExBAzwtAAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAAB"\
                     "AEAAABAAAAQAAAEA4DwA4SMAuAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAA"\
                     "AQBAAAAQAAAEAAABACAPgEfhAVwAAAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIA"\
                     "gAAAIAAACAAAAgCAAACwCYBH4QF6BuAzqhfMfvvtt9/+u/1+AgLoeQEIAIAAACAAAAgAAAIAgAAAIAAA"\
                     "CAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAAnAcgfAQAFwAAAgCAAAAgAAAIAAACAIAAACAA"\
                     "AAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAOQB8Cg8gAsAAAEAQAAAEAAABAAAAQBAAAAQ"\
                     "AAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAA2AfAoPEDPAKxRvWD222+//fbf7fcTEEDP"\
                     "C0AAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQDgPADhIwC4"\
                     "AAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAIA+AR+EB"\
                     "XAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAALAJgEfh"\
                     "AXoGYI3qBauten+n/fbbX3i/n4AAel4AAgAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAg"\
                     "AAAIAAACAIAAACAAAAgAAOcBCB8BwAUAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAA"\
                     "IAAACAAAAgCAAAAgAAAIAAB5ADwKD+ACAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAA"\
                     "ABAAAAQAAAEAQAAAEAAABACATQA8Cg/QMwCfUb1g9ttvv/323+33ExBAzwtAAAAEAAABAEAAABAAAAQA"\
                     "AAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEA4DwA4SMAuAAAEAAABAAAAQBAAAAQAAAE"\
                     "AAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABACAPgEfhAVwAAAgAAAIAgAAAIAAACAAA"\
                     "AgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAACwCYBH4QF6BmCN6gWz33777f+PKP/9"\
                     "/QQE0PMCEAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAADgP"\
                     "QPgIAC4AAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAADI"\
                     "A+BReAAXAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAA"\
                     "bALgUXiAlr7Y4BnEOVAUKwAAAABJRU5ErkJggg=="
            texfn = os.path.splitext(filepath)[0] + '.png'
            import base64
            with open(texfn, "wb") as fp:
                fp.write(base64.b64decode(palimg))
            mtl_line = _write_MTL(m, "palette_rgb", texture=texfn, opacity=obj["opacity"])
    elif m and "colour" in obj:
        #print("Writing mtl lib (default colour)")
        c = obj.parent.parse_colour(obj["colour"])
        mtl_line = _write_MTL(m, 'default-' + name, diffuse=c, opacity=obj["opacity"])

    for o,data in enumerate(obj):
        if verbose: print("[%s] element %d of %d" % (obj.name, o+1, len(obj.data.vertices)))
        verts = data.vertices.reshape((-1,3))
        if len(verts) == 0:
            if verbose: print("No vertices")
            continue
        f.write("o Surface_%d\n" % o)
        #f.write("o %s\n" % obj.name)
        if m: f.write("mtllib " + os.path.splitext(os.path.basename(filepath))[0] + ".mtl\n")
        indices = data.indices.reshape((-1,3))
        normals = data.normals.reshape((-1,3))
        texcoords = data.texcoords.reshape((-1,2))
        #Calculate texcoords from colour values?
        if len(texcoords) == 0 and obj["texture"] == 'colourmap':
            label = obj["colourby"]
            if isinstance(label,int):
                #Use the given label index
                sets = list(datasets.keys())
                label = sets[label]
            elif len(label) == 0:
                #Use the default label
                label = 'values'
            valdata = data[label]
            if len(valdata) >= o+1:
                #Found matching value array
                v = valdata
                #Normalise [0,1]
                texcoords = (v - numpy.min(v)) / numpy.ptp(v)
                #Add 2nd dimension (not actually necessary,
                #tex coords can by 1d but breaks some loaders (meshlab)
                zeros = numpy.zeros((len(texcoords)))
                texcoords = numpy.vstack((texcoords,zeros)).reshape([2, -1]).transpose()

        #Colours?
        cv0 = ""
        vperc = 1
        if len(data.colours):
            vperc = int(verts.shape[0] / len(data.colours))

        if verbose: print("- Writing vertices:",verts.shape)
        for vi,v in enumerate(verts):
            #Vertex colour with vertex? (only if flag passed)
            if vertexcolours and len(data.colours):
                #if vi%10==0: print(len(data.colours), len(verts), vperc, vi)
                c = data.colours[len(data.colours) // vperc - 1]
                rgb = colour2rgb(c)
                f.write("v %.6f %.6f %.6f %.6f %.6f %.6f\n" % (v[0], v[1], v[2], rgb[0]/255.0, rgb[1]/255.0, rgb[2]/255.0))
            else:
                f.write("v %.6f %.6f %.6f\n" % (v[0], v[1], v[2]))
        if verbose: print("- Writing normals:",normals.shape)
        for n in normals:
            f.write("vn %.6f %.6f %.6f\n" % (n[0], n[1], n[2]))
        if verbose: print("- Writing texcoords:",texcoords.shape)
        if len(texcoords.shape) == 2:
            for t in texcoords:
                f.write("vt %.6f %.6f\n" % (t[0], t[1]))
        else:
            for t in texcoords:
                f.write("vt %.6f\n" % t)

        #Face elements v/t/n v/t v//n
        f.write(mtl_line)
        if len(normals) and len(texcoords):
            if verbose: print("- Writing faces (v/t/n):",indices.shape)
        elif len(texcoords):
            if verbose: print("- Writing faces (v/t):",indices.shape)
        elif len(normals):
            if verbose: print("- Writing faces (v//n):",indices.shape)
        else:
            if verbose: print("- Writing faces (v):",indices.shape)
        if verbose: print("- Colours :",data.colours.shape)
        if verbose: print("- Indices :",indices.shape)

        for n,i in enumerate(indices):
            if verbose and n%1000==0: print(".", end=''); sys.stdout.flush()
            i0 = i[0]+offset
            i1 = i[1]+offset
            i2 = i[2]+offset
            #Use mtl colours?
            if colourdict:
                ci = int(i[0] / vperc)
                cv1 = data.colours[ci]
                if cv0 != cv1 and cv1 in colourdict:
                    f.write("usemtl " + colourdict[cv1][0] + "\n")
                    cv0 = cv1

            if len(normals) and len(texcoords):
                f.write("f %d/%d/%d %d/%d/%d %d/%d/%d\n" % (i0, i0, i0, i1, i1, i1, i2, i2, i2))
            elif len(texcoords):
                f.write("f %d/%d %d/%d %d/%d\n" % (i0, i0, i1, i1, i2, i2))
            elif len(normals):
                f.write("f %d//%d %d//%d %d//%d\n" % (i0, i0, i1, i1, i2, i2))
            else:
                f.write("f %d %d %d\n" % (i0, i1, i2))
        if verbose: print()

        offset += verts.shape[0]
    return offset

def export_PLY(filepath, source, binary=True):
    """
    Export given object(s) to a PLY file
    Supports points or triangle mesh object data

    If source is lavavu.Viewer() exports all objects
    If source is lavavu.Object() exports single object

    Parameters
    ----------
    filepath : str
        Output file to write
    source : lavavu.Viewer or lavavu.Object
        Where to get object data to export
    binary : boolean
        Write vertex/face data as binary, default True
    """
    objects = _get_objects(source)
    with open(filepath, mode='wb') as f:
        voffset = 0
        foffset = 0
        #First count vertices, faces
        fc = 0
        vc = 0
        for obj in objects:
            for o,data in enumerate(obj):
                vc += len(data.vertices) #Vertices now in shape (N,3) so len returns correct count
                fc += len(data.indices) // 3

        #Pass the counts first
        vertex = None
        face = None
        print(vc, " vertices, ", fc, " faces")
        for obj in objects:
            for o,data in enumerate(obj):
                print("[%s] element %d of %d, type %s" % (obj.name, o+1, len(obj.data), data.type))
                #print("OFFSETS:",voffset,foffset)
                verts = data.vertices.reshape((-1,3))
                if len(verts) == 0:
                    print("No vertices")
                    return
                if len(verts) != vc:
                    print("Vertex count error!", len(verts), vc, verts.shape)
                    return
                indices = data.indices.reshape((-1,3))
                normals = data.normals.reshape((-1,3))
                texcoords = data.texcoords.reshape((-1,2))

                vperc = 0
                cperf = 0
                if len(data.colours):
                    vperc = int(verts.shape[0] / len(data.colours))
                    C = data.colours
                    #print("COLOURS:",len(C),C.shape, verts.shape[0], verts.shape[0] / len(data.colours), vperc)

                if data.type != 'points':
                    #Per face colours, or less
                    if face is None:
                        if vperc and vperc < len(verts):
                            #cperf = int(indices.shape[0] / len(data.colours))
                            face = numpy.zeros(shape=(fc), dtype=[('vertex_indices', 'i4', (3,)), ('red', 'u1'), ('green', 'u1'), ('blue', 'u1')])
                        else:
                            face = numpy.zeros(shape=(fc), dtype=[('vertex_indices', 'i4', (3,))])
                        print("FACE:",face.dtype)

                    for i,idx in enumerate(indices):
                        if i%1000==0: print(".", end=''); sys.stdout.flush()
                        if vperc and vperc < len(verts):
                            #Have colour, but less than vertices, apply to faces
                            ci = idx[0] // vperc
                            c = data.colours[ci]
                            rgb = colour2rgb(c)
                            face[i+foffset] = ([idx[0]+voffset, idx[1]+voffset, idx[2]+voffset], rgb[0], rgb[1], rgb[2])
                        else:
                            face[i+foffset] = ([idx[0]+voffset, idx[1]+voffset, idx[2]+voffset])
                    print()

                #Construct and write vertex elements
                if vertex is None:
                    #Setup vertex array based on first object element
                    D = [('x', 'f4'), ('y', 'f4'), ('z', 'f4')]
                    if normals.shape[0] == verts.shape[0]:
                        D += [('nx', 'f4'), ('ny', 'f4'), ('nz', 'f4')]
                    if texcoords.shape[0] == verts.shape[0]:
                        D += [('s', 'f4'), ('t', 'f4')]
                    if vperc and vperc == 1:
                        D += [('red', 'u1'), ('green', 'u1'), ('blue', 'u1')]
                    print("VERTEX:",D)
                    vertex = numpy.zeros(shape=(vc), dtype=D)

                for i,v in enumerate(verts):
                    #if i%1000==0:
                    #    print("vert index",i,vperc)
                    if i%1000==0: print(".", end=''); sys.stdout.flush()
                    E = [v[0], v[1], v[2]]
                    if normals.shape[0] == verts.shape[0]:
                        N = normals[i]
                        E += [N[0], N[1], N[2]]
                    if texcoords.shape[0] == verts.shape[0]:
                        T = texcoords[i]
                        E += [T[0], T[1]]
                    if vperc and vperc == 1:
                        c = data.colours[i]
                        rgb = colour2rgb(c)
                        E += [rgb[0], rgb[1], rgb[2]]
                    vertex[i+voffset] = tuple(E)
                print()

                #Update offsets : number of vertices / faces added
                voffset += verts.shape[0]
                foffset += indices.shape[0]

        import plyfile
        #vertex = numpy.array(vertex, dtype=vertex.dtype)
        els = []
        els.append(plyfile.PlyElement.describe(vertex, 'vertex'))
        if face is not None:
            els.append(plyfile.PlyElement.describe(face, 'face'))

        #Write, text or binary
        if binary:
            print("Writing binary PLY data")
            plyfile.PlyData(els).write(f)
        else:
            print("Writing ascii PLY data")
            plyfile.PlyData(els, text=True).write(f)

def _get_PLY_colours(element):
    """
    Extract colour data from PLY element and return as a numpy rgba array
    """
    r = None
    g = None
    b = None
    a = None
    #print(element.properties)
    for prop in element.properties:
        #print(prop,prop.name,prop.dtype)
        if 'red' in prop.name: r = element[prop.name]
        if 'green' in prop.name: g = element[prop.name]
        if 'blue' in prop.name: b = element[prop.name]
        if 'alpha' in prop.name: a = element[prop.name]

    if r is not None and g is not None and b is not None:
        if a is None:
            a = numpy.full(r.shape, 255)
        #return numpy.array([r, g, b, a])
        return numpy.vstack((r,g,b,a)).reshape([4, -1]).transpose()

    return None

def plot_PLY(lv, filename):
    """
    Plot triangles from a PlyData instance. Assumptions:
    has a 'vertex' element with 'x', 'y', and 'z' properties,
    has a 'face' element with an integral list property 'vertex_indices',
    all of whose elements have length 3.
    """
    import plyfile
    plydata = plyfile.PlyData.read(filename)

    x = plydata['vertex']['x']
    y = plydata['vertex']['y']
    z = plydata['vertex']['z']
    V = numpy.vstack((x,y,z)).reshape([3, -1]).transpose()
    #V = numpy.array([x, y, z])
    #print("VERTS:", V.shape)

    vp = []
    for prop in plydata['vertex'].properties:
        vp.append(prop.name)
        print(prop.name)
    
    N = None
    if 'nx' in vp and 'ny' in vp and 'nz' in vp:
        nx = plydata['vertex']['nx']
        ny = plydata['vertex']['ny']
        nz = plydata['vertex']['nz']
        N = numpy.vstack((nx,ny,nz)).reshape([3, -1]).transpose()
    
    T = None
    if 's' in vp and 't' in vp:
        s = plydata['vertex']['s']
        t = plydata['vertex']['t']
        T = numpy.vstack((s,t)).reshape([2, -1]).transpose()

    C = _get_PLY_colours(plydata['vertex'])

    if 'face' in plydata:
        #print(plydata['face'])
        #Face colours?
        if C is None:
            C = _get_PLY_colours(plydata['face'])

        tri_idx = plydata['face']['vertex_indices']
        idx_dtype = tri_idx[0].dtype
        tri_idx = numpy.asarray(tri_idx).flatten()
        #print(type(tri_idx),idx_dtype,tri_idx.shape)
        #print(tri_idx)

        triangles = numpy.array([t.tolist() for t in tri_idx]).flatten()
        #triangles = numpy.fromiter(tri_idx, [('data', idx_dtype, (3,))], count=len(tri_idx))['data']

        return lv.triangles(vertices=V, indices=triangles, colours=C, normals=N, texcoords=T)
    else:
        return lv.points(vertices=V, colours=C, normals=N, texcoords=T)

def export_any(filepath, source, name=None):
    """
    Export given object(s) to a file format supproted by trimesh, eg: GLTF or GLB file
    See: https://trimsh.org/trimesh.exchange.html
    Requires "trimesh" module
    Supports triangle mesh object data

    If source is lavavu.Viewer() exports all objects
    If source is lavavu.Object() exports single object

    Parameters
    ----------
    filepath : str
        Output file to write
    source : lavavu.Viewer or lavavu.Object
        Where to get object data to export
    """
    #TODO: support points, lines
    try_import('trimesh')

    objects = _get_objects(source)
    scene = trimesh.Scene()
    facect = 0
    for obj in objects:
        #Use object name to export unless name provided
        oname = name
        if oname is None:
            oname = obj["name"]
        for i,e in enumerate(obj.data):
            #print(e)
            meshdict = {}
            meshdict["vertices"] = e.vertices.reshape(-1,3)
            meshdict["vertex_normals"] = e.normals.reshape(-1,3)
            meshdict["faces"] = e.indices.reshape(-1,3)
            facect += e.indices.shape[0]
            if not len(meshdict["faces"]):
                print("Empty",e)
                continue
            colours = e.colours
            #print("ColourCount",len(colours))
            if len(colours) == len(meshdict["vertices"]):
                #View int32 into uint8 bytes
                meshdict["vertex_colors"] = colours.view('uint8').reshape((-1,4))

            mesh = trimesh.load_mesh(meshdict)
            scene.add_geometry(mesh, geom_name=oname + '#' + str(i))

            if len(colours) <= 1:
                #If we don't set a default material colour, trimesh import will set default colour for every vertex
                colour = obj.parent.parse_colour(obj["colour"])
                colour = numpy.array(colour*255, dtype=numpy.uint8)
                if len(colours) > 0:
                    colour = colours.view('uint8').reshape((-1,4))[0]
                if obj["opacity"] < 1.0:
                    colour[3] *= obj["opacity"]
                #print("Single colour: export as material", colour)
                mesh.visual = trimesh.visual.TextureVisuals(material=trimesh.visual.material.PBRMaterial())
                mesh.visual.material.baseColorFactor = colour

    if facect > 0:
        scene.export(file_obj=filepath)
    else:
        print("No triangle facets to export")

    """
    #Try GLTF ascii?
    gltf = trimesh.exchange.gltf.export_gltf(scene)
    import json
    print("Mesh loaded, exporting GLTF")
    gltf = trimesh.exchange.gltf.export_gltf(mesh) #, merge_buffers=True)
    #print(gltf.keys())
    #print(gltf['model.gltf'])
    for fn in gltf:
        mode = 'w'
        #if ".bin" in fn: mode = 'wb'
        with open(fn, 'wb') as f:
            print("writing gltf component:" + fn)
            f.write(gltf[fn])
    """

def read_any(filepath, lv, scaling=None):
    """
    Load using trimesh, supports GLTF etc

    See: https://trimsh.org/trimesh.exchange.html
    Requires "trimesh" module
    """
    #TODO: support points, lines
    try_import('trimesh')
    scene = trimesh.load(filepath)
    tris = None
    for name in scene.geometry.keys():
        geometry = scene.geometry[name]
        idx = 0
        if '#' in name:
            name,idx = name.split('#')
            idx = int(idx)

        if idx == 0:
            #Texturing for multiple objects is broken when loaded with triangles() renderer
            #tris = lv.triangles(name)
            #Use surface() instead and textures work as expected
            tris = lv.surface(name)
        else:
            tris.append()

        if scaling:
            import numpy as np
            matrix = np.eye(4)
            matrix[0, 0] = scaling
            matrix[1, 1] = scaling
            matrix[2, 2] = scaling
            geometry.apply_transform(matrix)

        tris.vertices(geometry.vertices)
        tris.normals(geometry.vertex_normals)
        tris.indices(geometry.faces)

        #adjacency_matrix = geometry.edges_sparse
        #print(geometry)
        #print(geometry.visual)

        #Load vertex colours if available
        if hasattr(geometry.visual, "vertex_colors"):
            #print("HAVE VERTEX COLOURS",len(geometry.visual.vertex_colors))
            tris.colours(geometry.visual.vertex_colors)
        #Load single colour material
        elif hasattr(geometry.visual, "material"):
            if hasattr(geometry.visual.material, "baseColorTexture"):
                #print("HAVE TEXTURE")
                #print(geometry.visual.material)
                #if hasattr(geometry.visual, "image"):
                #print(geometry.visual.image)
                #print(dir(geometry.visual.material))
                #print(hasattr(geometry.visual, "material"))
                #print(geometry.visual.material.baseColorTexture)
                tris.texture(geometry.visual.material.baseColorTexture)
            elif hasattr(geometry.visual.material, "baseColorFactor"):
                #print("HAVE MATERIAL")
                #print(geometry.visual.material)
                #print(geometry.visual.material.baseColorFactor)
                if idx == 0:
                    #Can set as prop, but only works for single el
                    tris["colour"] = geometry.visual.material.baseColorFactor
                else:
                    tris.colours(geometry.visual.material.baseColorFactor)

        idx += 1 #Inc index?

    return tris



"""
Warning! EXPERIMENTAL:
 these features and functions are under development, will have bugs,
 and may be heavily modified in the future

Tracer particles in a vector field
Uses a KDTree to find nearest vector to advect the particles

- Requires scipy.spatial
"""
import numpy
import os
import sys
import random
import math
from scipy import spatial

#Necessary? for large trees, detect?
#sys.setrecursionlimit(10000)

class TracerState(object):
    def __init__(self, verts, N=5000):
        self.tree = spatial.cKDTree(verts)
        self.tracers = None
        self.steps = [0]*N
        self.values = None
        self.positions = None
        self.velocities = None

def trace_particles(state, verts, vecs, N=5000, limit=0.5, speed=1.0, noise=0.0, height=None):
    """
    Take a list of tracer vertices and matching velocity grid points (verts) & vectors (vecs)
    For each tracer

    - find the nearest velocity grid point
    - if within max dist: Multiply position by velocity vector
    - otherwise: Generate a new start position for tracer

    Parameters
    ----------
    state : TracerState
        Object returned from first call, pass None on first pass
    verts : array or list
        vertices of the vector field
    vecs : array or list
        vector values of the vector field
    N : int
        Number of particles to seed
    limit : float
        Distance limit over which tracers are not connected,
        For example if using a periodic boundary, setting limit to
        half the bounding box size will prevent tracer lines being
        connected when passing through the boundary
    speed : float
        Speed multiplier, scaling factor for the velocity taken from the vector values
    noise : float
        A noise factor, if set a random value is generated, multiplied by noise factor
        and added to each new position
    height : float
        A fixed height value, all positions will be given this height as their Z component

    Returns
    -------
    TracerState
        Object to hold the tracer state and track particles
        Pass this as first paramter on subsequent calls
    """

    #KDstate.tree for finding nearest velocity grid point
    if state is None:
        state = TracerState(verts, N)

    lastid = 0
    def rand_vert():
        #Get a random velocity grid point
        lastid = random.randint(0, len(verts)-1)
        pos = verts[lastid]
        #Generate some random noise to offset
        noise3 = numpy.array((0.,0.,0.))
        if noise > 0.0:
            noise3 = numpy.random.rand(3) * noise
        #Fixed height?
        if height:
            noise3[2] = height
        #Return the sum
        return pos + noise3

    if state.positions is None:
        state.positions = numpy.zeros(shape=(N,3))
        state.velocities = numpy.zeros(shape=(N,3))
        state.values = numpy.zeros(shape=(N))
        for i in range(N):
            state.positions[i] = rand_vert()
            state.steps[i] = 0
            state.values[i] = numpy.linalg.norm(vecs[lastid])

    #Query all tracer state.positions
    q = state.tree.query(state.positions, k=1)
    for r in range(len(q[0])):
        #print("result, distance, point")
        #print(r, q[0][r], state.tree.data[q[1][r]], state.positions[r])

        #Increasing random chance as steps exceed 5 of a new start pos
        if random.randint(0,state.steps[r]) > 5:
            #Pick a new random grid vertex to start from
            #(Must be farther away than distance criteria)
            old = numpy.array(state.positions[r])
            while True:
                state.positions[r] = rand_vert() 
                if numpy.linalg.norm(state.positions[r]-old) > limit:
                    break
            state.steps[r] = 0
            state.values[r] = numpy.linalg.norm(vecs[lastid])
        else:
            #Index of nearest grid point is in q[1][r]
            #Lookup vector at this index, multiply by position to get delta and add
            state.velocities[r] = vecs[q[1][r]] #Store velocity
            state.positions[r] += speed * vecs[q[1][r]]
            #Increment step tracking
            state.steps[r] += 1
            state.values[r] = numpy.linalg.norm(vecs[q[1][r]])

    return state

